<?php

namespace SGalinski\SgYoutube\Service;

/***************************************************************
 *  Copyright notice
 *
 *  (c) sgalinski Internet Services (https://www.sgalinski.de)
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use InvalidArgumentException;
use TYPO3\CMS\Core\Context\Context;
use TYPO3\CMS\Core\Context\Exception\AspectNotFoundException;
use TYPO3\CMS\Core\Context\LanguageAspect;
use TYPO3\CMS\Core\Http\ServerRequest;
use TYPO3\CMS\Core\Registry;
use TYPO3\CMS\Core\Site\Entity\Site;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\VersionNumberUtility;

/**
 * YouTube Helper Service
 */
class YoutubeService {
	const API_URL = 'https://www.googleapis.com/youtube/v3/';
	const API_CHANNEL = 'search';
	const API_PLAYLIST = 'playlistItems';
	const API_VIDEO = 'videos';
	const API_PART = 'snippet';
	const API_PART_LOCALIZATIONS = 'localizations';
	const API_ORDER_BY = 'date';

	/**
	 * Maps the json array from the YouTube call to return some unified value. The output from YouTube is pretty
	 * unsteady. Also we calculate the correct thumbnail sized and so on.
	 *
	 * @param array $jsonArray
	 * @param string $youtubeId
	 * @param string $aspectRatio 16:9 (default) or 4:3 (ONLY used if byAspectRation is set as thumbnail type)
	 * @param string $thumbnailType maxres, standard, high, medium, default, byAspectRatio (default)
	 * @param string $apiKey
	 * @return array
	 */
	public function mapArray(
		$jsonArray = [], $youtubeId = '', $aspectRatio = '16:9', $thumbnailType = 'byAspectRatio', $apiKey = ''
	): array {
		if (count($jsonArray) <= 0) {
			return $jsonArray;
		}

		// Normalize the data to video details.
		if (strpos($youtubeId, 'UC') === 0 || strpos($youtubeId, 'PL') === 0) {
			$result = $this->getDetailedVideoInformationForJsonArray($jsonArray, $apiKey, self::API_PART);
			if (count($result) <= 0 || !isset($result['items'])) {
				return $jsonArray;
			}

			$jsonArray = $result['items'];
		}

		if (!in_array($thumbnailType, ['maxres', 'standard', 'high', 'medium', 'default', 'byAspectRatio'])) {
			$thumbnailType = 'byAspectRatio';
		}

		if (!in_array($aspectRatio, ['16:9', '4:3'])) {
			$aspectRatio = '16:9';
		}

		// Localization is just available from TYPO3 9.X.X
		if (version_compare(VersionNumberUtility::getCurrentTypo3Version(), '9.0.0', '>=')) {
			$context = GeneralUtility::makeInstance(Context::class);

			try {
				/** @var LanguageAspect $languageAspect */
				$languageAspect = $context->getAspect('language');
				$currentLanguageUid = $languageAspect->getId();
			} catch (AspectNotFoundException $e) {
				// Can't be possible to land here, otherwise the whole frontend would be weird as hell..
				$currentLanguageUid = 0;
			}

			if ($currentLanguageUid > 0 && $youtubeId && $apiKey) {
				$jsonArray = $this->addLocalizationData($jsonArray, $apiKey, $currentLanguageUid);
			}
		}

		$result = [];
		foreach ($jsonArray as $field) {
			$previewImage = [];
			$resolutionTypes = ['maxres', 'standard', 'high', 'medium', 'default'];
			if ($thumbnailType !== 'byAspectRatio') {
				array_unshift($resolutionTypes, $thumbnailType);
			}

			foreach ($resolutionTypes as $type) {
				if (isset($field['snippet']['thumbnails'][$type])) {
					$previewImage = $field['snippet']['thumbnails'][$type];
					if ($thumbnailType === 'byAspectRatio' && isset($previewImage['height'])) {
						$aspectRatioOfImage = $previewImage['width'] / $previewImage['height'];
						if ($aspectRatio === '16:9' && $aspectRatioOfImage > 1.7 && $aspectRatioOfImage < 1.9) {
							break;
						}

						if ($aspectRatio === '4:3' && $aspectRatioOfImage > 1.2 && $aspectRatioOfImage < 1.4) {
							break;
						}
					} else {
						break;
					}
				}
			}

			// Don't cache the preview URL. YouTube has a CDN and delivers much faster.
			$result[] = [
				'title' => $field['snippet']['title'],
				'description' => strip_tags($field['snippet']['description']),
				'thumbnail' => $previewImage['url'],
				'url' => 'https://www.youtube.com/watch?v=' . $field['id']
			];
		}

		return $result;
	}

	/**
	 * Adds the localized title and description for each of the given entries in the jsonArray and returns it.
	 *
	 * @param array $jsonArray
	 * @param string $apiKey
	 * @param int $currentLanguageUid
	 *
	 * @return array
	 */
	protected function addLocalizationData(array $jsonArray, $apiKey, $currentLanguageUid): array {
		if (!$apiKey || !$currentLanguageUid || count($jsonArray) <= 0) {
			return $jsonArray;
		}

		$localizationData = $this->getDetailedVideoInformationForJsonArray(
			$jsonArray, $apiKey, self::API_PART_LOCALIZATIONS
		);
		if (!isset($localizationData['items']) || count($localizationData['items']) <= 0) {
			return $jsonArray;
		}

		/** @var ServerRequest $request */
		$request = $GLOBALS['TYPO3_REQUEST'];
		$attributes = $request->getAttributes();
		if (!isset($attributes['site'])) {
			return $jsonArray;
		}

		/** @var Site $site */
		$site = $attributes['site'];
		$languages = $site->getLanguages();
		$currentSiteLanguage = $languages[$currentLanguageUid];
		if (!$currentSiteLanguage) {
			return $jsonArray;
		}

		$languageIsoCodes = [
			$currentSiteLanguage->getTwoLetterIsoCode()
		];
		foreach ($currentSiteLanguage->getFallbackLanguageIds() as $languageId) {
			$siteLanguage = $languages[$languageId];
			if (!$siteLanguage) {
				continue;
			}

			$languageIsoCodes[] = $siteLanguage->getTwoLetterIsoCode();
		}

		foreach ($localizationData['items'] as $index => $localizationEntry) {
			if (!isset($localizationEntry['localizations']) || count($localizationEntry['localizations']) <= 0) {
				continue;
			}

			$title = '';
			$description = '';
			$localizations = $localizationEntry['localizations'];
			foreach ($languageIsoCodes as $languageIsoCode) {
				if ($title && $description) {
					break;
				}

				if (!$title &&  isset($localizations[$languageIsoCode]['title'])) {
					$title = $localizations[$languageIsoCode]['title'];
				}

				if (!$description &&  isset($localizations[$languageIsoCode]['description'])) {
					$description = $localizations[$languageIsoCode]['description'];
				}
			}

			if ($title) {
				$jsonArray[$index]['snippet']['title'] = $title;
			}

			if ($description) {
				$jsonArray[$index]['snippet']['description'] = $description;
			}
		}

		return $jsonArray;
	}

	/**
	 * Returns the detailed video information for the given json array and returns them as an array.
	 *
	 * @param array $jsonArray
	 * @param string $apiKey
	 * @param string $part
	 *
	 * @return array
	 */
	public function getDetailedVideoInformationForJsonArray(array $jsonArray, $apiKey, $part): array {
		if (!$apiKey || count($jsonArray) <= 0) {
			return $jsonArray;
		}

		$apiUrl = self::API_URL . self::API_VIDEO;
		$parameters = [];
		$parameters['part'] = $part;
		$parameters['key'] = $apiKey;
		$query = http_build_query($parameters);
		foreach ($jsonArray as $videoData) {
			$videoId = 0;
			if (isset($videoData['id'])) {
				$videoId = $videoData['id']['videoId'] ?? $videoData['id'];
			}

			if (!$videoId) {
				continue;
			}

			$query .= '&id=' . $videoId;
		}

		$result = $this->getJsonAsArray('', '10', $apiKey, $apiUrl . '?' . $query);
		if (!isset($result['items']) || count($result['items']) <= 0) {
			return $jsonArray;
		}

		return $result;
	}

	/**
	 * Returns a JSON array with the video details (title, description, preview image, url)
	 *
	 * @param string $youtubeId
	 * @param string $maxResults
	 * @param string $apiKey
	 * @param string $url
	 * @return array|mixed
	 */
	public function getJsonAsArray($youtubeId = '', $maxResults = '10', $apiKey = '', $url = '') {
		if (!$url) {
			$url = $this->getApiUrl($youtubeId, $maxResults, $apiKey);
		}

		$registry = GeneralUtility::makeInstance(Registry::class);
		$currentDay = date('Y-m-d', $GLOBALS['EXEC_TIME']);
		$cacheKey = sha1($url);
		$disableYoutubeCache = (bool) GeneralUtility::_GP('disableYoutubeCache');
		if (!$disableYoutubeCache) {
			$cachedResult = $registry->get('sg_youtube', $cacheKey);
			if ($cachedResult) {
				if ($cachedResult['CACHE_DATE'] === $currentDay) {
					return $cachedResult;
				}

				$registry->remove('sg_youtube', $cacheKey);
			}
		}

		if (function_exists('curl_init')) {
			$ch = curl_init();
			curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, TRUE);
			curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);
			curl_setopt($ch, CURLOPT_URL, $url);
			$jsonString = curl_exec($ch);
			curl_close($ch);

		} elseif (ini_get('allow_url_fopen')) {
			$jsonString = file_get_contents($url);
		} else {
			throw new InvalidArgumentException(
				'The system doesn\'t have CURL enabled and "allow_url_fopen" is disabled.', 403
			);
		}

		$jsonArray = ($jsonString !== '' ? json_decode($jsonString, TRUE) : []);
		if ($jsonArray === NULL) {
			throw new InvalidArgumentException(
				'There is something wrong with loaded JSON or encoded data is deeper than the recursion limit.', 403
			);
		}

		if (!$jsonArray) {
			throw new InvalidArgumentException('JSON could\'t be parsed.', 403);
		}

		// By API documentation provided 'error' key is sent if, probably,
		// URL can not return JSON data or Permission is denied.
		if (isset($jsonArray['error'])) {
			throw new InvalidArgumentException('Message: ' . $jsonArray['error']['message'], 403);
		}

		if (!isset($jsonArray['items'])) {
			throw new InvalidArgumentException('No items array.', 403);
		}

		if (count($jsonArray['items']) < 1) {
			throw new InvalidArgumentException('No items found.', 403);
		}

		if (!$disableYoutubeCache) {
			$jsonArray['CACHE_DATE'] = $currentDay;
			$registry->set('sg_youtube', $cacheKey, $jsonArray);
		}

		return $jsonArray;
	}

	/**
	 * Returns the YouTube API URL
	 *
	 * @param string $youtubeId
	 * @param string $maxResults
	 * @param string $apiKey
	 * @return string
	 */
	public function getApiUrl($youtubeId = '', $maxResults = '10', $apiKey = ''): string {
		$apiUrl = self::API_URL;
		$parameters = [];

		if (strpos($youtubeId, 'UC') === 0) {
			$apiUrl .= self::API_CHANNEL;
			$parameters['channelId'] = $youtubeId;
		} elseif (strpos($youtubeId, 'PL') === 0) {
			$apiUrl .= self::API_PLAYLIST;
			$parameters['playlistId'] = $youtubeId;
		} else {
			$apiUrl .= self::API_VIDEO;
			$parameters['id'] = $this->removeIdParameters($youtubeId);
		}

		$parameters['order'] = self::API_ORDER_BY;
		$parameters['part'] = self::API_PART;
		$parameters['key'] = $apiKey;
		$parameters['maxResults'] = $maxResults;
		$query = http_build_query($parameters);

		return $apiUrl . '?' . $query;
	}

	/**
	 * Removes GET parameters following the ID
	 *
	 * @param string $youtubeId
	 * @return array|string
	 */
	protected function removeIdParameters($youtubeId = '') {
		if (strpos($youtubeId, '&')) {
			return explode('&', $youtubeId)[0];
		}

		return $youtubeId;
	}
}
